from unittest2 import TestCase
from describe import expect

from exam.decorators import fixture


class Outer(object):

    @classmethod
    def meth(cls):
        return cls, 'from method'


class Dummy(object):

    outside = 'value from outside'

    @fixture
    def number(self):
        return 42

    @fixture
    def obj(self):
        return object()

    inline = fixture(int, 5)
    inline_obj = fixture(object)
    inline_func = fixture(lambda self: self.outside)
    inline_func_with_args = fixture(lambda *a, **k: (a, k), 1, 2, a=3)
    inline_from_method = fixture(Outer.meth)


class TestFixture(TestCase):

    def test_converts_method_to_property(self):
        expect(Dummy().number).to == 42

    def test_caches_property_on_same_instance(self):
        instance = Dummy()
        expect(instance.obj).to.be_equal_to(instance.obj)

    def test_gives_different_object_on_separate_instances(self):
        expect(Dummy().obj).to_not.be_equal_to(Dummy().obj)

    def test_can_be_used_inline_with_a_function(self):
        expect(Dummy().inline_func).to == 'value from outside'

    def test_can_be_used_with_a_callable_that_takes_args(self):
        inst = Dummy()
        expect(inst.inline_func_with_args).to == ((inst, 1, 2), dict(a=3))

    def test_can_be_used_with_class_method(self):
        expect(Dummy().inline_from_method).to == (Outer, 'from method')

    def test_if_passed_type_builds_new_object(self):
        expect(Dummy().inline).to == 5



from __future__ import absolute_import

from mock import patch
from functools import partial
import types


class fixture(object):

    def __init__(self, thing, *args, **kwargs):
        self.thing = thing
        self.args = args
        self.kwargs = kwargs

    def __get__(self, instance, type=None):
        if not self.thing in instance.__dict__:
            applied = self.__apply(instance)(*self.args, **self.kwargs)
            instance.__dict__[self.thing] = applied

        return instance.__dict__[self.thing]

    def __apply(self, instance):
        if type(self.thing) in (type, types.MethodType):
            return self.thing
        else:
            return partial(self.thing, instance)


class base(object):

    def __init__(self, thing):
        self.thing = thing

    def __call__(self, instance):
        return self.thing(instance)


class before(base):
    pass


class after(base):
    pass


class around(base):
    pass


class patcher(object):

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.func = None

    def __call__(self, func):
        self.func = func
        return self

    def build_patch(self, instance):
        if self.func:
            self.kwargs['new'] = self.func(instance)

        return patch(*self.args, **self.kwargs)



import inspect

from decorators import before, after, around, patcher


class MultipleGeneratorsContextManager(object):

    def __init__(self, *generators):
        self.generators = generators

    def __enter__(self, *args, **kwargs):
        [g.next() for g in self.generators]

    def __exit__(self, *args, **kwargs):
        for generator in self.generators:
            try:
                generator.next()
            except StopIteration:
                pass


class Exam(object):

    @before
    def setup_patchers(self):
        for attr, patchr in self.attrs_of_type(patcher):
            patch_object = patchr.build_patch(self)
            setattr(self, attr, patch_object.start())
            self.addCleanup(patch_object.stop)

    def attrs_of_type(self, kind):
        for base in inspect.getmro(type(self)):
            for attr, value in vars(base).items():
                if type(value) is kind:
                    yield attr, value

    def setUp(self):
        for _, value in self.attrs_of_type(before):
            value(self)

    def tearDown(self):
        for _, value in self.attrs_of_type(after):
            value(self)

    def run(self, *args, **kwargs):
        generators = (value(self) for _, value in self.attrs_of_type(around))
        with MultipleGeneratorsContextManager(*generators):
            super(Exam, self).run(*args, **kwargs)



from unittest2 import TestCase
from describe import expect

from exam.objects import no_op


class TestNoOp(TestCase):

    def test_can_be_called_with_anything(self):
        no_op()
        no_op(1)
        no_op(key='val')
        no_op(1, key='val')
        no_op(1, 2, 3, key='val')
        no_op(1, 2, 3, key='val', another='thing')

    def test_returns_none(self):
        expect(no_op()).to == None



from mock import sentinel, MagicMock
from unittest2 import TestCase
import itertools

from exam.decorators import fixture, before, after, around, patcher
from exam.cases import Exam

from describe import expect

from dummy import get_thing, get_it


class FakeTest(object):

    def __init__(self):
        self.cleanups = []

    def run(self, *args, **kwargs):
        self.state_when_run = list(self.calls)

    def addCleanup(self, func):
        self.cleanups.append(func)


class DummyTest(Exam, FakeTest):

    @patcher('tests.dummy.thing')
    def dummy_thing(self):
        return sentinel.mock

    dummy_it = patcher('tests.dummy.it', return_value=12)

    def __init__(self):
        self.calls = []
        super(DummyTest, self).__init__()

    @before
    def append_one(self):
        self.calls.append(1)

    @after
    def append_two(self):
        self.calls.append(2)

    @around
    def append_5_then_6(self):
        self.calls.append(5)
        yield
        self.calls.append(6)


class ExtendedDummy(DummyTest):

    @before
    def append_3(self):
        self.calls.append(3)

    @after
    def append_4(self):
        self.calls.append(4)

    @around
    def append_7_then_8(self):
        self.calls.append(7)
        yield
        self.calls.append(8)


# TODO: Make the subclass checking just be a subclass of the test case
class TestExam(Exam, TestCase):

    @fixture
    def case(self):
        return DummyTest()

    @fixture
    def subclass_case(self):
        return ExtendedDummy()

    @after
    def stop_patchers(self):
        cleanups = (self.case.cleanups, self.subclass_case.cleanups)

        for cleanup in itertools.chain(*cleanups):
            if hasattr(cleanup.im_self, 'is_local'):  # Is the mock started?
                cleanup()

    @property
    def other_thing(self):
        return get_thing()

    @property
    def other_it(self):
        return get_it()

    def test_before_adds_each_method_to_set_up(self):
        expect(self.case.calls).to == []
        self.case.setUp()
        expect(self.case.calls).to == [1]

    def test_after_adds_each_method_to_tear_down(self):
        expect(self.case.calls).to == []
        self.case.tearDown()
        expect(self.case.calls).to == [2]

    def test_around_calls_methods_before_and_after_run(self):
        expect(self.case.calls).to == []
        self.case.run()
        expect(self.case.state_when_run).to == [5]
        expect(self.case.calls).to == [5, 6]

    def test_before_works_on_subclasses(self):
        expect(self.subclass_case.calls).to == []
        self.subclass_case.setUp()
        expect(self.subclass_case.calls).to == [3, 1]

    def test_after_works_on_subclasses(self):
        expect(self.subclass_case.calls).to == []
        self.subclass_case.tearDown()
        expect(self.subclass_case.calls).to == [4, 2]

    def test_around_works_with_subclasses(self):
        expect(self.subclass_case.calls).to == []
        self.subclass_case.run()
        expect(self.subclass_case.state_when_run).to == [7, 5]
        expect(self.subclass_case.calls).to == [7, 5, 8, 6]

    def test_patcher_start_value_is_added_to_case_dict_on_setup(self):
        self.case.setUp()
        expect(self.case.dummy_thing).to == sentinel.mock

    def test_patcher_patches_object_on_setup_and_adds_patcher_to_cleanup(self):
        expect(self.other_thing).to != sentinel.mock
        self.case.setUp()
        expect(self.other_thing).to == sentinel.mock
        [cleanup() for cleanup in self.case.cleanups]
        expect(self.other_thing).to != sentinel.mock

    def test_patcher_lifecycle_works_on_subclasses(self):
        expect(self.other_thing).to != sentinel.mock
        self.subclass_case.setUp()
        expect(self.other_thing).to == sentinel.mock
        [cleanup() for cleanup in self.subclass_case.cleanups]
        expect(self.other_thing).to != sentinel.mock

    def test_patcher_patches_with_a_magic_mock_if_no_function_decorated(self):
        expect(self.other_it()).to != 12
        self.case.setUp()
        expect(self.other_it()).to == 12
        self.case.cleanups[0]()
        expect(self.other_thing).to != 12



#!/usr/bin/env python
import sys

from setuptools import setup, find_packages

try:
    import multiprocessing
except ImportError:
    pass

install_requires = ['mock']
lint_requires = ['pep8', 'pyflakes']
tests_require = ['nose', 'unittest2', 'describe==1.0.0beta1']

dependency_links = [
    'https://github.com/jeffh/describe/tarball/dev#egg=describe'
]

setup_requires = []
if 'nosetests' in sys.argv[1:]:
    setup_requires.append('nose')

setup(
    name='exam',
    version='0.4.1',
    author='Jeff Pollard',
    author_email='jeff.pollard@gmail.com',
    url='https://github.com/fluxx/exam',
    description='Helpers for better testing.',
    license='MIT',
    packages=find_packages(),
    install_requires=install_requires,
    tests_require=tests_require,
    setup_requires=setup_requires,
    extras_require={
        'test': tests_require,
        'all': install_requires + tests_require,
        'docs': ['sphinx'] + tests_require,
        'lint': lint_requires
    },
    zip_safe=False,
    test_suite='nose.collector',
)